-- ScriptEnv.lua

local function LocalScriptEnv()
	return Battle.local_script_env_
end

local ScriptEnv = {}
ScriptEnv.Easing = CoroutineUtility.EasingFunctions

-- functions
function ScriptEnv.Damage(animation_name, se_name)
	ScriptEnv.Reaction()

	local env = LocalScriptEnv()

	if rawget(env, 'avoid') then
		return
	end

	-- guard
	if rawget(env, 'guard') then
		animation_name = nil
		ScriptEnv.CreateEffect('ef_all_guard')
			:position(env.target:position() + Vector3.New(0.0, 1.1, -0.5))
		ScriptEnv.PlaySE('guard')
	end

	-- se & animation
	ScriptEnv.PlaySE(se_name)
	if animation_name ~= nil then
		env.target:PlayAnimation(animation_name)
	end
	if not rawget(env, 'guard') then
		ScriptEnv.CallCharacterEvent(env.target, 'OnDamage')
	end
end
function ScriptEnv.Reaction()
	local env = LocalScriptEnv()

	if rawget(env, 'used_reaction_') then
		return
	end
	rawset(env, 'used_reaction_', true)

	if rawget(env, 'avoid') then
		env.target:PlayAnimation('avoid')
		ScriptEnv.PlaySE('avoid')
		ScriptEnv.CallCharacterEvent(env.target, 'OnAvoid')
	elseif rawget(env, 'guard') then
		if not env.continuous_attack then
			env.target:PlayAnimation('guard')
			ScriptEnv.CallCharacterEvent(env.target, 'OnGuard')
		end
	end
end

function ScriptEnv.CallCharacterEvent(character, event)
	local func = character[event]
	if func ~= nil then
		func(character)
	end
end

-- Wait
function ScriptEnv.Wait(time)
	while time > 0.0 do
		time = time - coroutine.yield()
	end
end

-- thread
function ScriptEnv.CreateThread(func)
	return ThreadList.New(LocalScriptEnv().thread_list_:CreateThread(func))
end
function ScriptEnv.WaitThread(thread)
	if thread == nil then
		thread = LocalScriptEnv().thread_list_
	end

	while not thread:IsCompleted() do
		coroutine.yield()
	end
end

-- disposable objects
local function DisposeObject(self, ...)
	if self.this_object_ ~= nil then
		local obj = self.this_object_
		self.this_object_ = nil
		obj[self.dispose_function_name_](obj, ...)
	end
end
local function UnregisterAndDisposeObject(self, ...)
	ScriptEnv.UnregisterDisposableObject(self)
	DisposeObject(self, ...)
end
function ScriptEnv.RegisterDisposableObject(obj, function_name)
	local new_disposable_object = {
		this_object_ = obj,
		dispose_function_name_ = function_name
	}
	new_disposable_object[function_name] = UnregisterAndDisposeObject
	setmetatable(new_disposable_object, {__index = obj})
	table.insert(LocalScriptEnv().disposable_object_list_, new_disposable_object)
	return new_disposable_object
end
function ScriptEnv.UnregisterDisposableObject(obj)
	local list = LocalScriptEnv().disposable_object_list_
	for i, value in ipairs(list) do
		if value == obj then
			table.remove(list, i)
			break
		end
	end
end
function ScriptEnv.DisposeAllRegistedObject()
	local env = LocalScriptEnv()
	local list = env.disposable_object_list_
	for i, value in ipairs(list) do
		DisposeObject(value)
	end
	env.disposable_object_list_ = {}
end

-- audio
function ScriptEnv.PlaySE(name)
	Battle.System:PlaySE(name)
end

-- effect
local function LoadEffect(name)
	Battle.System:LoadEffect(name)
end
function ScriptEnv.LoadEffect(name)
	if type(name) ~= 'table' then
		LoadEffect(name)
	else
		for i, value in ipairs(name) do
			LoadEffect(value)
		end
	end
end
function ScriptEnv.CreateEffect(name)
	local effect = Battle.System:SpawnEffect(name)
	local attacker_rotation = rawget(LocalScriptEnv(), 'attacker_rotation')
	if attacker_rotation ~= nil then
		effect:rotation(attacker_rotation)
	end
	return ScriptEnv.RegisterDisposableObject(effect, 'Destroy')
end
function ScriptEnv.CreateHitEffect()
	local env = LocalScriptEnv()
	if not rawget(env, 'avoid') then
		local hit_effect_names = {'ef_all_hit_small', 'ef_all_hit_large', 'ef_all_hit_minimum'}
		local name = nil
		local range_compatibility = rawget(env, 'range_compatibility')
		if range_compatibility ~= nil then
			name = hit_effect_names[1 + range_compatibility]
		else
			name = hit_effect_names[2]
		end
		return ScriptEnv.CreateEffect(name)
	else
		return NilObject.New()
	end
end

-- camera
local function LoadCameraAnimation(name)
	Battle.System:LoadCameraAnimation(name)
end
function ScriptEnv.LoadCameraAnimation(name)
	if type(name) ~= 'table' then
		LoadCameraAnimation(name)
	else
		for i, value in ipairs(name) do
			LoadCameraAnimation(value)
		end
	end
end
function ScriptEnv.PlayCameraAnimation(name)
	Battle.System:PlayCameraAnimation(name)
end
--[[
function ScriptEnv.WaitCameraAnimation()
	while Battle.System:CameraAnimationIsPlaying() do
		coroutine.yield()
	end
end
]]

-- background
function ScriptEnv.SetBackgroundActive(active)
	local env = LocalScriptEnv()
	if active then
		local backbround_deactive = rawget(env, 'backbround_deactive_')
		if backbround_deactive ~= nil then
			backbround_deactive:Reset()
		end
	else
		if rawget(env, 'backbround_deactive_') == nil then
			Battle.System:SetBackgroundActive(false)
			local t = { Reset = function(self) Battle.System:SetBackgroundActive(true) rawset(env, 'backbround_deactive_', nil) end }
			t = ScriptEnv.RegisterDisposableObject(t, 'Reset')
			rawset(env, 'backbround_deactive_', t)
		end
	end
end

-- metatable
setmetatable(ScriptEnv, { __index = _G })
setmetatable(_G, {__index = function() return NilObject.New() end})

return ScriptEnv
