-- ScriptEnv.lua

local function LocalScriptEnv()
	return Battle.local_script_env_
end

local ScriptEnv = {}
ScriptEnv.Easing = CoroutineUtility.EasingFunctions

-- functions
function ScriptEnv.GetParties()
	return Battle.System:GetParties()
end
function ScriptEnv.SetHitCount(count)
	local env = LocalScriptEnv()
	if env.hit_count_ == nil then
		rawset(env, 'hit_count_', count)
		rawset(env, 'hit_', 0)
	end
end
function ScriptEnv.Damage(animation_name, se_name)
	ScriptEnv.Reaction()

	local env = LocalScriptEnv()

	if env.avoid then
		return
	end

	-- damage
	local rate = 1.0
	if env.hit_count_ == nil then
		ScriptEnv.SetHitCount(1)
	end
	if env.hit_ >= env.hit_count_ then
		return
	end
	env.hit_ = env.hit_ + 1
	if env.hit_ < env.hit_count_ then
		rate = env.hit_ / env.hit_count_
	end
	Battle.System:Damage(rate)

	-- guard
	if env.guard then
		animation_name = nil
		ScriptEnv.CreateEffect('ef_all_guard')
			:position(env.target:position() + Vector3.New(0.0, 1.1, -0.5))
		ScriptEnv.PlaySE('guard')
	end

	-- finish
	local finish = false
	if rate >= 1.0 then
		if not env.finished_ then
			rawset(env, 'finished_', true)
			if env.finish or env.badgebroken then
				ScriptEnv.PlaySE('finish')
				if not env.meteor_art then
					animation_name = 'finish'
					ScriptEnv.PlayCameraAnimation('camera_finish')
					finish = true
				end
				if env.finish then
					ScriptEnv.CreateAnimationObject('Unconsciousness')
				elseif env.badgebroken then
					ScriptEnv.CreateAnimationObject('BadgeBroken')
				end
			end
		end
	end

	-- se & animation
	ScriptEnv.PlaySE(se_name)
	if animation_name ~= nil then
		env.target:PlayAnimation(animation_name)
	end
	if not env.guard then
		ScriptEnv.CallCharacterEvent(env.target, 'OnDamage')
	end

	-- finish time scale
	if finish then
		ScriptEnv.SetTimeScale(0.0, 2.5)
		while coroutine.yield() <= 0.0 do end
		ScriptEnv.SetTimeScale(0.03, 2.0)
	end
end
function ScriptEnv.Reaction()
	local env = LocalScriptEnv()

	if env.used_reaction_ then
		return
	end
	rawset(env, 'used_reaction_', true)

	if env.avoid then
		env.target:PlayAnimation('avoid')
		ScriptEnv.PlaySE('avoid')
		ScriptEnv.CallCharacterEvent(env.target, 'OnAvoid')
	elseif env.guard then
		if not env.continuous_attack then
			env.target:PlayAnimation('guard')
			ScriptEnv.CallCharacterEvent(env.target, 'OnGuard')
		end
	end
end

function ScriptEnv.Regenerate()
	Battle.System:Regenerate()
end
function ScriptEnv.MiasmaDamage()
	Battle.System:MiasmaDamage()
end

function ScriptEnv.CallCharacterEvent(character, event)
	local func = character[event]
	if func ~= nil then
		func(character)
	end
end

-- Wait
function ScriptEnv.Wait(time)
	while time > 0.0 do
		time = time - coroutine.yield()
	end
end

-- thread
function ScriptEnv.CreateThread(func)
	return ThreadList.New(LocalScriptEnv().thread_list_:CreateThread(func))
end
function ScriptEnv.WaitThread(thread)
	if thread == nil then
		thread = LocalScriptEnv().thread_list_
	end

	while not thread:IsCompleted() do
		coroutine.yield()
	end
end

-- disposable objects
local function DisposeObject(self, ...)
	if self.this_object_ ~= nil then
		local obj = self.this_object_
		self.this_object_ = nil
		obj[self.dispose_function_name_](obj, ...)
	end
end
local function UnregisterAndDisposeObject(self, ...)
	ScriptEnv.UnregisterDisposableObject(self)
	DisposeObject(self, ...)
end
function ScriptEnv.RegisterDisposableObject(obj, function_name)
	local new_disposable_object = {
		this_object_ = obj,
		dispose_function_name_ = function_name
	}
	new_disposable_object[function_name] = UnregisterAndDisposeObject
	setmetatable(new_disposable_object, {__index = obj})
	table.insert(LocalScriptEnv().disposable_object_list_, new_disposable_object)
	return new_disposable_object
end
function ScriptEnv.UnregisterDisposableObject(obj)
	local list = LocalScriptEnv().disposable_object_list_
	for i, value in ipairs(list) do
		if value == obj then
			table.remove(list, i)
			break
		end
	end
end
function ScriptEnv.DisposeAllRegistedObject()
	local env = LocalScriptEnv()
	local list = env.disposable_object_list_
	for i, value in ipairs(list) do
		DisposeObject(value)
	end
	env.disposable_object_list_ = {}
end

-- sprite
function ScriptEnv.CreateSprite(name)
	local sprite = Battle.System:CreateSprite(name)
	return ScriptEnv.RegisterDisposableObject(sprite, 'Destroy')
end

-- animation objects
function ScriptEnv.CreateAnimationObject(name)
	local animation_object = Battle.System:CreateAnimationObject(name)
	return ScriptEnv.RegisterDisposableObject(animation_object, 'Destroy')
end
function ScriptEnv.CreateTitleAnimationObject()
	local animation_object = Battle.System:CreateTitleAnimationObject()
	return ScriptEnv.RegisterDisposableObject(animation_object, 'Destroy')
end
function ScriptEnv.CreateResultWinObject()
	local animation_object = Battle.System:CreateResultWinObject()
	return ScriptEnv.RegisterDisposableObject(animation_object, 'Destroy')
end

-- audio
function ScriptEnv.PlaySE(name)
	Battle.System:PlaySE(name)
end

-- effect
local function LoadEffect(name)
	Battle.System:LoadEffect(name)
end
function ScriptEnv.LoadEffect(name)
	if type(name) ~= 'table' then
		LoadEffect(name)
	else
		for i, value in ipairs(name) do
			LoadEffect(value)
		end
	end
end
function ScriptEnv.CreateEffect(name)
	local effect = Battle.System:SpawnEffect(name)
	local env = LocalScriptEnv()
	if env.attacker_rotation ~= nil then
		effect:rotation(env.attacker_rotation)
	end
	return ScriptEnv.RegisterDisposableObject(effect, 'Destroy')
end
function ScriptEnv.CreateHitEffect()
	local env = LocalScriptEnv()
	if not env.avoid then
		local hit_effect_names = {'ef_all_hit_small', 'ef_all_hit_large', 'ef_all_hit_minimum'}
		local name = nil
		if env.range_compatibility ~= nil then
			name = hit_effect_names[1 + env.range_compatibility]
		else
			name = hit_effect_names[2]
		end
		return ScriptEnv.CreateEffect(name)
	else
		return NilObject.New()
	end
end

-- camera
local function LoadCameraAnimation(name)
	Battle.System:LoadCameraAnimation(name)
end
function ScriptEnv.LoadCameraAnimation(name)
	if type(name) ~= 'table' then
		LoadCameraAnimation(name)
	else
		for i, value in ipairs(name) do
			LoadCameraAnimation(value)
		end
	end
end
function ScriptEnv.PlayCameraAnimation(name, offset)
	local env = LocalScriptEnv()
	Battle.System:PlayCameraAnimation(name, env.animation_camera)
	if offset == nil then
		offset = Vector3.zero()
	end
	offset = Battle.SetPositionConversion(offset)
	Battle.System:SetCameraPositionOffset(env.animation_camera, offset)
end
function ScriptEnv.WaitCameraAnimation()
	while Battle.System:CameraAnimationIsPlaying() do
		coroutine.yield()
	end
end

-- time scale
function ScriptEnv.SetTimeScale(time_scale, time, callback)
	if time_scale == nil then
		return
	end
	if time == nil then
		time = 0.0
	end

	ScriptEnv.ResetTimeScale()

	if time_scale ~= 1.0 then
		local env = LocalScriptEnv()
		local t = {
			Reset = function(self)
				env.time_scale_ = nil
				Battle.System:ResetTimeScale()
				if callback ~= nil then
					callback()
				end
			end
		}
		t = ScriptEnv.RegisterDisposableObject(t, 'Reset')
		rawset(env, 'time_scale_', t)
		Battle.System:SetTimeScale(time_scale, time, function()
			if env.time_scale_ ~= nil then
				env.time_scale_:Reset()
			end
		end)
	else
		if callback ~= nil then
			callback()
		end
	end
end
function ScriptEnv.ResetTimeScale()
	local env = LocalScriptEnv()
	if env.time_scale_ ~= nil then
		env.time_scale_:Reset()
	end
end

function ScriptEnv.SetCameraTimeScale(time_scale)
	if time_scale == nil then
		return
	end

	ScriptEnv.ResetCameraTimeScale()

	if time_scale ~= 1.0 then
		local env = LocalScriptEnv()
		local t = {
			Reset = function(self)
				env.camera_time_scale_ = nil
				Battle.System:SetCameraTimeScale(1.0)
			end
		}
		t = ScriptEnv.RegisterDisposableObject(t, 'Reset')
		rawset(env, 'camera_time_scale_', t)
		Battle.System:SetCameraTimeScale(time_scale)
	end
end
function ScriptEnv.ResetCameraTimeScale()
	local env = LocalScriptEnv()
	if env.camera_time_scale_ ~= nil then
		env.camera_time_scale_:Reset()
	end
end

function ScriptEnv.SetTimeScaleWithCameraTimeScale(time_scale, time)
	ScriptEnv.SetCameraTimeScale(time_scale)
	ScriptEnv.SetTimeScale(time_scale, time, function() ScriptEnv.ResetCameraTimeScale() end)
end

-- background
function ScriptEnv.SetBackgroundActive(active)
	local env = LocalScriptEnv()
	if active then
		if env.backbround_deactive_ ~= nil then
			env.backbround_deactive_:Reset()
		end
	else
		if env.backbround_deactive_ == nil then
			Battle.System:SetBackgroundActive(false)
			local t = { Reset = function(self) Battle.System:SetBackgroundActive(true) env.backbround_deactive_ = nil end }
			t = ScriptEnv.RegisterDisposableObject(t, 'Reset')
			rawset(env, 'backbround_deactive_', t)
		end
	end
end

-- metatable
setmetatable(ScriptEnv, { __index = _G })

return ScriptEnv
