-- Vector3.lua

local Vector3 = {}
local Vector3_metatable = {}
setmetatable(Vector3, Vector3_metatable)
local Vector3_instance_functions = {}
local Vector3_instance_metatable = { __index = Vector3_instance_functions }

-- static
function Vector3.New(x, y, z)
	local instance = {
		x = x or 0.0,
		y = y or 0.0,
		z = z or 0.0
	}
	setmetatable(instance, Vector3_instance_metatable)
	return instance
end

function Vector3.IsInherited(t)
	return Utility.IsInherited(t, Vector3_instance_metatable)
end

function Vector3.Dot(lhs, rhs)
	return lhs.x * rhs.x + lhs.y * rhs.y + lhs.z * rhs.z
end

function Vector3.Cross(lhs, rhs)
	return Vector3.New(
		lhs.y * rhs.z - lhs.z * rhs.y,
		lhs.z * rhs.x - lhs.x * rhs.z,
		lhs.x * rhs.y - lhs.y * rhs.x)
end

function Vector3.zero()
	return Vector3.New(0.0, 0.0, 0.0)
end

function Vector3.one()
	return Vector3.New(1.0, 1.0, 1.0)
end

function Vector3.Scale(lhs, rhs)
	return Vector3.New(lhs.x * rhs.x, lhs.y * rhs.y, lhs.z * rhs.z)
end

function Vector3.Lerp(v1, v2, t)
	return Vector3.New(
		v1.x + (v2.x - v1.x) * t,
		v1.y + (v2.y - v1.y) * t,
		v1.z + (v2.z - v1.z) * t)
end

function Vector3.QuadraticBezier(v1, v2, v3, t)
	local rt = 1.0 - t
	local rt2 = rt * rt
	local _2rt_t = 2 * rt * t
	local t2 = t * t
	return Vector3.New(
		rt2 * v1.x + _2rt_t * v2.x + t2 * v3.x,
		rt2 * v1.y + _2rt_t * v2.y + t2 * v3.y,
		rt2 * v1.z + _2rt_t * v2.z + t2 * v3.z)
end
function Vector3.CubicBezier(v1, v2, v3, v4, t)
	local rt = 1.0 - t
	local rt3 = rt * rt * rt
	local _3_rt2_t = 3 * rt * rt * t
	local _3_rt_t2 = 3 * rt * t * t
	local t3 = t * t * t
	return Vector3.New(
		rt3 * v1.x + _3_rt2_t * v2.x + _3_rt_t2 * v3.x + t3 * v4.x,
		rt3 * v1.y + _3_rt2_t * v2.y + _3_rt_t2 * v3.y + t3 * v4.y,
		rt3 * v1.z + _3_rt2_t * v2.z + _3_rt_t2 * v3.z + t3 * v4.z)
end

-- metatable
function Vector3_metatable.__call(t, ...)
	return Vector3.New(...)
end

-- instance
function Vector3_instance_functions:Clone()
	return Vector3.New(self.x, self.y, self.z)
end

function Vector3_instance_functions:Normalize()
	local reverse_magnitude = 1.0 / self:Magnitude()
	self.x = self.x * reverse_magnitude
	self.y = self.y * reverse_magnitude
	self.z = self.z * reverse_magnitude
end

function Vector3_instance_functions:Normalized()
	local clone = self:Clone()
	clone:Normalize()
	return clone
end

function Vector3_instance_functions:Magnitude()
	return math.sqrt(self:SqrMagnitude())
end

function Vector3_instance_functions:SqrMagnitude()
	return self.x * self.x + self.y * self.y + self.z * self.z
end

function Vector3_instance_functions:ToString()
	return '('..self.x..','..self.y..','..self.z..')'
end

-- metatable
function Vector3_instance_metatable.__add(lhs, rhs)
	return Vector3.New(lhs.x + rhs.x, lhs.y + rhs.y, lhs.z + rhs.z)
end
function Vector3_instance_metatable.__sub(lhs, rhs)
	return Vector3.New(lhs.x - rhs.x, lhs.y - rhs.y, lhs.z - rhs.z)
end
function Vector3_instance_metatable.__mul(lhs, rhs)
	return Vector3.New(lhs.x * rhs, lhs.y * rhs, lhs.z * rhs)
end

return Vector3
