-- Vector4.lua

local Vector4 = {}
local Vector4_metatable = {}
setmetatable(Vector4, Vector4_metatable)
local Vector4_instance_functions = {}
local Vector4_instance_metatable = { __index = Vector4_instance_functions }

-- static
function Vector4.New(x, y, z, w)
	local instance = {
		x = x or 0.0,
		y = y or 0.0,
		z = z or 0.0,
		w = w or 0.0
	}
	setmetatable(instance, Vector4_instance_metatable)
	return instance
end

function Vector4.IsInherited(t)
	return Utility.IsInherited(t, Vector4_instance_metatable)
end

function Vector4.Dot(lhs, rhs)
	return lhs.x * rhs.x + lhs.y * rhs.y + lhs.z * rhs.z + lhs.w * rhs.w
end

function Vector4.Lerp(v1, v2, t)
	return Vector4.New(
		v1.x + (v2.x - v1.x) * t,
		v1.y + (v2.y - v1.y) * t,
		v1.z + (v2.z - v1.z) * t,
		v1.w + (v2.w - v1.w) * t)
end

function Vector4.Plane(normal, point)
	return Vector4.New(normal.x, normal.y, normal.z, Vector3.Dot(normal, point))
end

-- metatable
function Vector4_metatable.__call(t, ...)
	return Vector4.New(...)
end

-- instance
function Vector4_instance_functions:Clone()
	return Vector4.New(self.x, self.y, self.z, self.w)
end

function Vector4_instance_functions:Scale(scale)
	return Vector4.New(self.x * scale.x, self.y * scale.y, self.z * scale.z, self.w * scale.w)
end

function Vector4_instance_functions:Normalize()
	local reverse_magnitude = 1.0 / self:Magnitude()
	self.x = self.x * reverse_magnitude
	self.y = self.y * reverse_magnitude
	self.z = self.z * reverse_magnitude
	self.w = self.w * reverse_magnitude
end

function Vector4_instance_functions:Magnitude()
	return math.sqrt(self:SqrMagnitude())
end

function Vector4_instance_functions:SqrMagnitude()
	return Vector4.Dot(self, self)
end

function Vector4_instance_functions:ToString()
	return '('..self.x..','..self.y..','..self.z..','..self.w..')'
end

function Vector4_instance_functions:ReflectionVector3(point)
	return Vector3.New(
		-- x
		(1.0 - 2.0 * self.x * self.x) * point.x +
		(    - 2.0 * self.x * self.y) * point.y + 
		(    - 2.0 * self.x * self.z) * point.z + 
		(    - 2.0 * self.x * self.w),
		-- y
		(    - 2.0 * self.y * self.x) * point.x +
		(1.0 - 2.0 * self.y * self.y) * point.y + 
		(    - 2.0 * self.y * self.z) * point.z + 
		(    - 2.0 * self.y * self.w),
		-- z
		(    - 2.0 * self.z * self.x) * point.x +
		(    - 2.0 * self.z * self.y) * point.y + 
		(1.0 - 2.0 * self.z * self.z) * point.z + 
		(    - 2.0 * self.z * self.w))
end

-- metatable
function Vector4_instance_metatable.__add(lhs, rhs)
	return Vector4.New(lhs.x + rhs.x, lhs.y + rhs.y, lhs.z + rhs.z, lhs.w + rhs.w)
end
function Vector4_instance_metatable.__sub(lhs, rhs)
	return Vector4.New(lhs.x - rhs.x, lhs.y - rhs.y, lhs.z - rhs.z, lhs.w - rhs.w)
end
function Vector4_instance_metatable.__mul(lhs, rhs)
	return Vector4.New(lhs.x * rhs, lhs.y * rhs, lhs.z * rhs, lhs.w * rhs)
end

return Vector4
